# -*- coding: UTF-8 -*-
"""
This module provides classes for storing drilling data in a database.

.. todo:: - reformat docstrings, especially of setter and getter functions
          - check exception types
"""

import sqlalchemy as sq
from sqlalchemy.orm import relationship
from sqlalchemy.orm.session import Session
from typing import List

from GeologicalToolbox.Exceptions import DatabaseException, WellMarkerDepthException
from GeologicalToolbox.GeoObject import AbstractGeoObject
from GeologicalToolbox.Geometries import GeoPoint
from GeologicalToolbox.WellLogs import WellLog
from GeologicalToolbox.DBHandler import Base, AbstractDBObject
from GeologicalToolbox.Stratigraphy import StratigraphicObject


class WellMarker(Base, AbstractDBObject):
    """
    Class WellMarker

    Represents single markers in a drilled well
    """
    # define db table name and columns
    __tablename__ = 'well_marker'

    id = sq.Column(sq.INTEGER, sq.Sequence('wellmarker_id_seq'), primary_key=True)
    drill_depth = sq.Column(sq.FLOAT)

    # define relationship to stratigraphic table
    horizon_id = sq.Column(sq.INTEGER, sq.ForeignKey('stratigraphy.id'))
    hor = relationship("StratigraphicObject")

    well_id = sq.Column(sq.INTEGER, sq.ForeignKey('wells.id'), default=-1)

    def __init__(self, depth, horizon, *args, **kwargs):
        # type: (float, StratigraphicObject, *object, **object) -> None
        """
        Creates a new well marker

        :param depth: depth below Kelly Bushing
        :type depth: float

        :param horizon: stratigraphy object
        :type horizon: StratigraphicObject

        :param args: parameters for AbstractDBObject initialisation
        :type args: List()

        :param kwargs: parameters for AbstractDBObject initialisation
        :type kwargs: Dict()

        :returns: Nothing
        :raises ValueError: Raises ValueError if one of the committed parameters cannot be converted to the expected type
        """
        AbstractDBObject.__init__(self, *args, **kwargs)
        if (type(horizon) is not StratigraphicObject) and (horizon is not None):
            raise ValueError("'horizon' value is not of type StratigraphicObject!")

        self.depth = float(depth)
        self.horizon = horizon

    def __repr__(self):
        # type: () -> str
        """
        Returns a text-representation of the well marker

        :return: Returns a text-representation of the well marker
        :rtype: str
        """
        return "<WellMarker(id='{}', depth='{}', horizon='{}'\nAbstractObject: {}". \
            format(self.id, self.depth, self.horizon, AbstractDBObject.__repr__(self))

    def __str__(self):
        # type: () -> str
        """
        Returns a text-representation of the well

        :return: Returns a text-representation of the well
        :rtype: str
        """
        return "[{}] {}: {} - {}".format(self.id, self.depth, self.horizon, AbstractDBObject.__str__(self))

    @property
    def depth(self):
        # type: () -> float
        """
        Returns the depth of the well marker below Kelly Bushing.

        :return: Returns the depth of the well marker below Kelly Bushing.
        :rtype: float
        """
        return float(self.drill_depth)

    @depth.setter
    def depth(self, value):
        # type: (float) -> None
        """
        Sets a new depth below Kelly Bushing for the well marker

        :param value: depth below Kelly Bushing
        :type value: float

        :return: Nothing
        :raises ValueError: Raises ValueError if value if not of type float or cannot be converted to float
        """
        self.drill_depth = float(value)

    @property
    def horizon(self):
        # type: () -> StratigraphicObject
        """
        Returns the stratigraphy of the point

        :return: Returns the current Stratigraphy
        :rtype: StratigraphicObject
        """
        return self.hor

    @horizon.setter
    def horizon(self, value):
        # type: (StratigraphicObject) -> None
        """
        sets a new stratigraphy

        :param value: new stratigraphy
        :type value: StratigraphicObject or None

        :return: Nothing

        :raises TypeError: Raises TypeError if value is not of type Stratigraphy
        """
        if (value is not None) and (type(value) is not StratigraphicObject):
            raise TypeError('type of commited value ({}) is not StratigraphicObject!'.format(type(value)))

        if value is None:
            self.hor = None
            self.horizon_id = -1
        else:
            self.hor = value

    def to_geopoint(self):
        # type: () -> GeoPoint
        """
        Returns the current well marker as GeoPoint

        :return: Returns the current well marker as GeoPoint
        :rtype: GeoPoint
        """
        easting = self.well.easting
        northing = self.well.northing
        altitude = self.well.altitude - self.depth
        return GeoPoint(self.horizon, True, self.well.reference_system, easting, northing, altitude, self.session,
                        self.well.well_name, self.comment)

    # load points from db
    @classmethod
    def load_in_extent_from_db(cls, session, min_easting, max_easting, min_northing, max_northing):
        # type: (Session, float, float, float, float) -> List[WellMarker]
        """
        Returns all well marker with committed horizon inside the given extent in the database connected to the
        SQLAlchemy Session session

        :param min_easting: minimal easting of extent
        :type min_easting: float

        :param max_easting: maximal easting of extent
        :type max_easting: float

        :param min_northing: minimal northing of extent
        :type min_northing: float

        :param max_northing: maximal northing of extent
        :type max_northing: float

        :param session: represents the database connection as SQLAlchemy Session
        :type session: Session

        :return: a list of wells representing the result of the database query
        :rtype: List[WellMarker]
        """
        result = session.query(cls, Well). \
            filter(Well.id == cls.well_id). \
            filter(sq.between(Well.east, min_easting, max_easting)). \
            filter(sq.between(Well.north, min_northing, max_northing)). \
            order_by(cls.id).all()
        for marker in result:
            marker.session = session
        return result

    @classmethod
    def load_all_by_stratigraphy_from_db(cls, horizon, session):
        # type: (StratigraphicObject) -> List[WellMarker]
        """
        Returns all WellMarker in the database which are connected to the horizon 'horizon'

        :param horizon: stratigraphy for the database query
        :type horizon: StratigraphicObject

        :param session: represents the database connection as SQLAlchemy Session
        :type session: Session

        :return: a list of WellMarker
        :rtype: List[WellMarker]
        """
        return session.query(cls).filter(WellMarker.horizon_id == horizon.id)

    @classmethod
    def load_all_by_stratigraphy_in_extent_from_db(cls, horizon, min_easting, max_easting, min_northing, max_northing,
                                                   session):
        # type: (StratigraphicObject) -> List[WellMarker]
        """
        Returns all WellMarker in the database which are connected to the horizon 'horizon'

        :param horizon: stratigraphy for the database query
        :type horizon: StratigraphicObject

        :param min_easting: minimal easting of extent
        :type min_easting: float

        :param max_easting: maximal easting of extent
        :type max_easting: float

        :param min_northing: minimal northing of extent
        :type min_northing: float

        :param max_northing: maximal northing of extent
        :type max_northing: float

        :param session: represents the database connection as SQLAlchemy Session
        :type session: Session

        :return: a list of WellMarker
        :rtype: List[WellMarker]
        """
        result = session.query(cls, Well). \
            filter(cls.horizon_id == horizon.id). \
            filter(Well.id == cls.well_id). \
            filter(sq.between(Well.east, min_easting, max_easting)). \
            filter(sq.between(Well.north, min_northing, max_northing)). \
            order_by(cls.id).all()
        for marker in result:
            marker.session = session
        return result


class Well(Base, AbstractGeoObject):
    """
    Class Well

    Represents a well for storage of geoscientific data.
    """
    # define db table name and columns
    __tablename__ = 'wells'

    id = sq.Column(sq.INTEGER, sq.Sequence('well_id_seq'), primary_key=True)
    drill_depth = sq.Column(sq.FLOAT)
    wellname = sq.Column(sq.VARCHAR(100), unique=True)
    shortwellname = sq.Column(sq.VARCHAR(100), default="")

    # define markers relationship
    marker = relationship("WellMarker", order_by=WellMarker.drill_depth,
                          backref="well", primaryjoin='Well.id == WellMarker.well_id',
                          cascade="all, delete, delete-orphan")

    logs = relationship("WellLog", order_by=WellLog.id,
                        backref="well", primaryjoin='Well.id == WellLog.well_id',
                        cascade="all, delete, delete-orphan")

    sq.Index('coordinate_index', AbstractGeoObject.east, AbstractGeoObject.north)

    def __init__(self, well_name, short_name, depth, *args, **kwargs):
        # type: (str, str, float, *object, **object) -> None
        """
        Creates a new Well

        :param session: SQLAlchemy session, which includes the database connection
        :type session: Session

        :param well_name: well name
        :type well_name: str

        :param short_name: shortened well name
        :type short_name: str

        :param depth: drilled depth of the well
        :type depth: float

        :param args: parameters for AbstractGeoObject initialisation
        :type args: List()

        :param kwargs: parameters for AbstractGeoObject initialisation
        :type kwargs: Dict()

        :return: Nothing
        :raises ValueError: Raises ValueError if one of the types cannot be converted
        """

        self.depth = depth
        self.well_name = well_name
        self.short_name = short_name

        # call base class constructor
        AbstractGeoObject.__init__(self, *args, **kwargs)

    def __repr__(self):
        # type: () -> str
        """
        Returns a text-representation of the well

        :return: Returns a text-representation of the well
        :rtype: str
        """
        return "<Well(id='{}', name='{}', short_name='{}', depth='{}', marker='{}', logs='{}' +" \
               "'comment='{}', marker='{}')>".format(self.id, self.name, self.short_name, self.easting, self.northing,
                                                     self.altitude, self.depth, self.comment, repr(self.marker))

    def __str__(self):
        # type: () -> str
        """
        Returns a text-representation of the well

        :return: Returns a text-representation of the well
        :rtype: str
        """
        text = "[{}] {} ({}):\n{} - {} - {} - {} - {}" \
            .format(self.id, self.name, self.short_name, self.easting, self.northing, self.altitude, self.depth,
                    self.comment)

        for marker in self.marker:
            text += "\n" + str(marker)

        return text

    @property
    def depth(self):
        # type: () -> float
        """
        Returns the drilling depth of the well

        :return: Returns the drilling depth of the well
        :rtype: float
        """
        return float(self.drill_depth)

    @depth.setter
    def depth(self, dep):
        # type: (float) -> None
        """
        Sets a new drilling depth for the well

        :param dep: New drilling depth
        :type dep: float

        :return: Nothing
        :raises ValueError: Raises ValueError if dep is not of type float or cannot be converted to float or when
                            depth is < 0
        """
        dep = float(dep)
        if dep < 0:
            raise ValueError('Depth is below 0! ({})'.format(dep))
        if (len(self.marker) > 0) and (dep < self.marker[-1].depth):
            raise WellMarkerDepthException(
                    'New depth ({}) lower than depth of last marker {}'.format(dep, self.marker[-1].depth))
        self.drill_depth = dep

    @property
    def well_name(self):
        # type: () -> str
        """
        Returns the name of the well

        :return: Returns the name of the well
        :rtype: str
        """
        return self.wellname

    @well_name.setter
    def well_name(self, name):
        # type: (str) -> None
        """
        Sets a new name for the well with a maximum of 100 characters

        :param name: well name
        :type name: str

        :return: Nothing
        """
        name = str(name)
        if len(name) > 100:
            name = name[:100]
        self.wellname = name

    @property
    def short_name(self):
        # type: () -> str
        """
        Returns the short name of the well

        :return: Returns the short name of the well
        :rtype: str
        """
        return self.shortwellname

    @short_name.setter
    def short_name(self, short_name):
        # type: (str) -> None
        """
        Sets a new short name for the well with a maximum of 20 characters

        :param short_name: well name
        :type short_name: str

        :return: Nothing
        """
        short_name = str(short_name)
        if len(short_name) > 20:
            short_name = short_name[:20]
        self.shortwellname = short_name

    def insert_marker(self, marker):
        # type: (WellMarker) -> None
        """
        Insert a new WellMarker in the well
        ATTENTION: If you insert a marker, the well will be automatically stored in the database!

        :param marker: WellMarker to be inserted
        :type marker: WellMarker

        :return: Nothing

        :raises TypeError: Raises TypeError if marker is not of type WellMarker
        :raises ValueError: Raises ValueError if the depth of the marker is larger than the drilled depth of the well
        """
        if type(marker) is not WellMarker:
            raise TypeError('marker {} is not of type WellMarker!'.format(str(marker)))
        if marker.depth > self.depth:
            raise ValueError('Marker depth ({}) is larger than final well depth ({})!'.format(marker.depth, self.depth))
        self.marker.append(marker)

        # new sorting to ensure correct order without storage and reloading from the database
        self.marker.sort(key=lambda x: x.depth)

    def insert_multiple_marker(self, marker):
        # type: (List[WellMarker]) -> None
        """
        Insert the multiple marker in the well
        ATTENTION: If you insert marker, the well will be automatically stored in the database!

        :param marker: List of marker to be inserted
        :type marker: List[WellMarker]

        :return: Nothing

        :raises TypeError: Raises TypeError if one of the marker is not of type WellMarker
        :raises ValueError: Raises ValueError if the depth of a marker is larger than the drilled depth of the well
        """
        for mark in marker:
            if type(mark) is not WellMarker:
                raise TypeError(
                        'At least on marker is not of type WellMarker ({}: {})!'.format(str(mark), str(type(mark))))
            if mark.depth > self.depth:
                raise ValueError('Marker depth ({}) is larger than final well depth ({})!'.
                                 format(mark.depth, self.depth))

        self.marker += marker

        # new sorting to ensure correct order without storage and reloading from the database
        self.marker.sort(key=lambda x: x.depth)

    def get_marker_by_depth(self, depth):
        # type: (float) -> WellMarker or None
        """
        Returns the marker at depth 'depth'

        :param depth: depth of the requested marker
        :type depth: float

        :return: Returns the marker at depth 'depth'
        :rtype: WellMarker

        :raises ValueError: Raises ValueError if no marker was found for the committed depth or depth is
                            not compatible to float
        """
        depth = float(depth)
        for marker in self.marker:
            if marker.depth == depth:
                return marker
        raise ValueError('No marker found at depth {}'.format(depth))

    def delete_marker(self, marker):
        # type: (WellMarker) -> None
        """
        Deletes the marker from the well

        :param marker: WellMarker object which should be deleted
        :type marker: WellMarker

        :return: Nothing

        :raises TypeError: Raises TypeError if marker is not of type WellMarker
        :raises ValueError: Raises ValueError the marker is not part of the well
        """
        if type(marker) is not WellMarker:
            raise TypeError('marker {} is not of type WellMarker!'.format(str(marker)))

        try:
            self.marker.remove(marker)
        except ValueError as e:
            raise ValueError(str(e) + '\nWellMarker with ID ' + str(marker.id) + ' not found in list!')

    def add_log(self, log):
        # type: (WellLog) -> None
        """
        Adds a new log to the well

        :param log: new well log
        :type log: WellLog

        :return: Nothing
        :raises TypeError: Raises TypeError if log is not of type WellLog
        """
        if type(log) is not WellLog:
            raise TypeError('log {} is not of type WellLog!'.format(str(log)))

        self.logs.append(log)

    def delete_log(self, log):
        # type: (WellLog) -> None
        """
        Deletes a log from the well

        :param log: log to delete
        :type log: WellLog

        :return: Nothing
        :raises TypeError: Raises TypeError if log is not of type WellLog
        :raises ValueError: Raises ValueError if log is not part of self.logs
        """
        if type(log) is not WellLog:
            raise TypeError('log {} is not of type WellLog!'.format(str(log)))

        try:
            self.logs.remove(log)
        except ValueError as e:
            raise ValueError(str(e) + '\nWellLog with ID ' + str(log.id) + ' not found in list!')

    @classmethod
    def load_by_wellname_from_db(cls, name, session):
        # type: (str, Session) -> Well
        """
        Returns the well with the given name in the database connected to the SQLAlchemy Session session

        :param name: Only the well with this name will be returned
        :type name: str

        :param session: represents the database connection as SQLAlchemy Session
        :type session: Session

        :return: As the name is a unique value, only one result can be returned or None
        :rtype: Well or None

        :raises DatabaseException: Raises DatabaseException if more than one result was found (name is an unique value)
        """
        result = session.query(cls).filter(cls.wellname == name)
        if result.count() == 0:
            return None
        if result.count() == 1:
            result = result.one()
            result.session = session
            return result

        raise DatabaseException('More than one ({}) well with the same name: {}! Database error!'.
                                format(result.count(), name))

    @classmethod
    def load_deeper_than_value_from_db(cls, session, min_depth):
        # type: (Session, float) -> List[Well]
        """
        Returns all wells with a drilled depth below the min_depth in the database connected to the SQLAlchemy Session
        session

        :param min_depth: minimal drilled depth
        :type min_depth: float

        :param session: represents the database connection as SQLAlchemy Session
        :type session: Session

        :return: a list of wells representing the result of the database query
        :rtype: List[Well]
        """
        result = session.query(cls).filter(cls.drill_depth >= min_depth)
        result = result.order_by(cls.id).all()
        for well in result:
            well.session = session
        return result



"""
This module hosts the class Requests, which provides functionality for special (geo-)database requests.
"""

import sqlalchemy as sq
from sqlalchemy.orm.session import Session
from typing import List, Tuple
import sys

from GeologicalToolbox.Exceptions import DatabaseException, DatabaseRequestException, FaultException, ListOrderException
from GeologicalToolbox.Geometries import GeoPoint
from GeologicalToolbox.Properties import Property
from GeologicalToolbox.Stratigraphy import StratigraphicObject
from GeologicalToolbox.Wells import WellMarker


class Requests:
    """
    The class Requests, which provides functionality for special (geo-)database requests.
    """

    def __init__(self):
        # type: () -> None
        """
        Initialise the class

        -> Currently nothing to do...
        """
        pass

    @staticmethod
    def check_extent(extent):
        # type: (List[float, float, float, float]) -> None
        """
        checks, if the given extent has the right format

        :param extent: value to be checked
        :type extent:
        :return: Nothing
        :raises TypeError: if extent is not a list
        :raises ValueError: if on list element is not compatible to float or number of elements is not 4
        :raises ListOrderException: if the ordering of the extent list [min_easting, max_easting, min_northing,
                max_northing] is wrong.
        """
        if extent is None:
            return

        if not isinstance(extent, List):
            raise TypeError("extent is not an instance of List()!")
        if len(extent) != 4:
            raise ValueError("Number of extension list elements is not 4!")
        for i in range(len(extent)):
            try:
                extent[i] = float(extent[i])
            except ValueError as e:
                if sys.version_info[0] == 2:
                    raise ValueError('At least on extent element cannot be converted to float!\n' + e.message)
                elif sys.version_info[0] == 3:
                    raise ValueError('At least on extent element cannot be converted to float!\n{}'.format(e))
                else:
                    raise ValueError('At least on extent element cannot be converted to float!')
        if extent[0] > extent[1]:
            raise ListOrderException("min easting > max easting")
        if extent[2] > extent[3]:
            raise ListOrderException("min northing > max northing")

    @staticmethod
    def create_thickness_point(sorted_dict, well_id, marker_1, marker_2, session, use_faulted=False, fault_name='',
                               add_properties=tuple()):
        # type: (dict, int, int, int, Session, bool, str, Tuple) -> GeoPoint
        """
        Generate a new GeoPoint with thickness property from 2 well marker

        :param sorted_dict: dictionary containing well_id / WellMarker data
        :type sorted_dict: dict()
        :param well_id: current well_id
        :type well_id: int
        :param marker_1: id of marker 1
        :type marker_1: int
        :param marker_2: id of marker 2
        :type marker_2: int
        :param session: current SQLAlchemy session
        :type session: Session
        :param use_faulted: should faulted sequence be included?
        :type use_faulted: bool
        :param fault_name: name of fault stratigraphic unit (default: "Fault")
        :type fault_name: str
        :param add_properties: Adds the properties to the GeoPoint. Format for each property: (name, unit, value)
        :type add_properties: tuple()
        :return: new GeoPoint Object
        :rtype: GeoPoint
        :raises FaultException: if a fault is inside the section and use_faulted is False
        :raises ValueError: if a property in the add_property tuple has less than 3 entries
        """

        min_depth = sorted_dict[well_id][marker_1].depth
        max_depth = sorted_dict[well_id][marker_2].depth

        faults = session.query(WellMarker).join(StratigraphicObject). \
            filter(WellMarker.horizon_id == StratigraphicObject.id). \
            filter(StratigraphicObject.unit_name == fault_name). \
            filter(WellMarker.well_id == well_id)
        if min_depth > max_depth:
            faults = faults.filter(sq.between(WellMarker.drill_depth, max_depth, min_depth))
        else:
            faults = faults.filter(sq.between(WellMarker.drill_depth, min_depth, max_depth))
        if (faults.count() > 0) and (use_faulted is False):
            raise FaultException("Fault inside section")

        point = sorted_dict[well_id][marker_1].to_geopoint()
        thickness = Property('thickness', 'm', session)
        thickness.value = max_depth - min_depth
        point.add_property(thickness)
        if use_faulted:
            faulted = Property('faulted', 'bool', session)
            if faults.count() > 0:
                faulted.value = 1
            else:
                faulted.value = 0
            point.add_property(faulted)
        for prop in add_properties:
            if len(prop) < 3:
                raise ValueError('property tuple has less than 3 entries!')
            new_property = Property(prop[0], prop[1], session)
            new_property.value = prop[2]
            point.add_property(new_property)
        return point

    @staticmethod
    def well_markers_to_thickness(session, marker_1, marker_2, *args, **kwargs):
        # type: (Session, str, str, *object, **object) -> List[GeoPoint]
        """
        This static method generates a point set including a thickness property derived from the committed well marker

        :param session: The SQLAlchemy session connected to the database storing the geodata
        :type session: Session
        :param marker_1: First WellMarker unit name
        :type marker_1: str
        :param marker_2: Second WellMarker unit name
        :type marker_2: str
        :param summarise_multiple: Summarise multiple occurrences of marker_1 and marker_2 to a maximum thickness. If this parameter is False (default value) create multiple points.
        :type summarise_multiple: bool
        :param extent: extension rectangle as list which borders the well distribution. The list has the following order: [min easting, max easting, min northing, max northing]
        :type extent: [float, float, float, float]
        :param use_faulted: if True, also sections with faults between marker_1 and marker_2 are returned
        :type use_faulted: bool
        :param fault_name: unit name of fault marker (standard: 'Fault')
        :type fault_name: str
        :return: A list of GeoPoints each with a thickness property
        :rtype: [GeoPoint]
        :raises AttributeError: if marker_1 and marker_2 are equal
        :raises DatabaseException: if the database query results in less than 2 marker of a well_id
        :raises DatabaseRequestException: if an unexpected query result occurs
        :raises TypeError: if session is not an instance of SQLAlchemy session
        :raises ValueError: if a parameter is not compatible with the required type

        for further raises see :meth:`Requests.check_extent`

        Query for selecting markers:

        .. code-block:: sql
            :linenos:

            SELECT wm1.* FROM well_marker wm1
            JOIN stratigraphy st1
            ON wm1.horizon_id = st1.id
            WHERE st1.unit_name IN ("mu", "so")
            AND EXISTS
            (
                SELECT 1 FROM well_marker wm2
                JOIN stratigraphy st2
                ON wm2.horizon_id = st2.id
                WHERE st2.unit_name IN ("mu", "so")
                AND wm1.well_id = wm2.well_id
                AND st1.unit_name <> st2.unit_name
            )
            ORDER BY wm1.well_id,wm1.drill_depth
        """
        if not isinstance(session, Session):
            raise TypeError("session is not of type SQLAlchemy Session")

        if marker_1 == marker_2:
            raise AttributeError('marker_1 and marker_2 cannot be equal!')

        summarise = True
        extent = None
        use_faulted = False
        fault_name = 'Fault'
        for i in range(len(args)):
            if i == 0:
                summarise = bool(args[i])
            elif i == 1:
                Requests.check_extent(args[i])
                extent = args[i]
            elif i == 2:
                use_faulted = bool(args[i])
            elif i == 3:
                fault_name = str(args[i])

        for key in kwargs:
            if key == 'summarise_multiple':
                summarise = bool(kwargs[key])
            elif key == 'extent':
                Requests.check_extent(kwargs[key])
                extent = kwargs[key]
            elif key == 'use_faulted':
                use_faulted = bool(kwargs[key])
            elif key == 'fault_name':
                fault_name = str(kwargs[key])

        result = session.query(WellMarker)
        if extent is None:
            statement = sq.text("SELECT wm1.* FROM well_marker wm1 " +
                                "JOIN stratigraphy st1 ON wm1.horizon_id = st1.id " +
                                "WHERE st1.unit_name IN (:marker1, :marker2) " +
                                "AND EXISTS " +
                                "( SELECT 1 FROM well_marker wm2 JOIN stratigraphy st2 ON wm2.horizon_id = st2.id " +
                                "WHERE st2.unit_name IN (:marker1, :marker2) AND wm1.well_id = wm2.well_id " +
                                "AND st1.unit_name <> st2.unit_name) ORDER BY wm1.well_id,wm1.drill_depth")
            result = result.from_statement(statement). \
                params(marker1=marker_1, marker2=marker_2). \
                all()
        else:
            statement = sq.text("SELECT wm1.* FROM well_marker wm1 " +
                                "JOIN wells ON wm1.well_id = wells.id " +
                                "JOIN stratigraphy st1 ON wm1.horizon_id = st1.id " +
                                "WHERE wells.east BETWEEN :east_min AND :east_max " +
                                "AND wells.north BETWEEN :north_min AND :north_max " +
                                "AND st1.unit_name IN (:marker1, :marker2)" +
                                "AND EXISTS " +
                                "( SELECT 1 FROM well_marker wm2 JOIN stratigraphy st2 ON wm2.horizon_id = st2.id " +
                                "WHERE st2.unit_name IN (:marker1, :marker2) AND wm1.well_id = wm2.well_id " +
                                "AND st1.unit_name <> st2.unit_name) ORDER BY wm1.well_id,wm1.drill_depth")
            result = result.from_statement(statement). \
                params(marker1=marker_1, marker2=marker_2, east_min=extent[0], east_max=extent[1], north_min=extent[2],
                       north_max=extent[3]). \
                all()

        # first: sort by well_id for simpler multiple marker check
        sorted_dict = dict()
        for i in range(len(result)):
            if result[i].well_id not in sorted_dict:
                sorted_dict[result[i].well_id] = list()
            sorted_dict[result[i].well_id].append(result[i])

        del result

        # generate the resulting list of GeoPoints
        geopoints = list()
        for well_id in sorted_dict:
            if len(sorted_dict[well_id]) < 2:
                raise DatabaseException('Not enough well marker in dictionary')
            if len(sorted_dict[well_id]) == 2:
                sorted_dict[well_id][0].session = session
                try:
                    if summarise:
                        point = Requests.create_thickness_point(sorted_dict, well_id, 0, 1, session, use_faulted,
                                                                fault_name, (('summarised', 'bool', 0),))
                        geopoints.append(point)
                    else:
                        point = Requests.create_thickness_point(sorted_dict, well_id, 0, 1, session, use_faulted,
                                                                fault_name, (('multiple marker', 'bool', 0),))
                        geopoints.append(point)
                # FaultException -> do nothing except catching the exception
                except FaultException:
                    continue
                # don't test anything else for this well_id
                continue

            # last case: more than 2 values found:
            if summarise:
                first_index = -1
                last_index = -1
                for i in range(len(sorted_dict[well_id])):
                    if sorted_dict[well_id][i].horizon.unit_name == marker_1:
                        first_index = i
                        break
                for j in reversed(range(len(sorted_dict[well_id]))):
                    if sorted_dict[well_id][j].horizon.unit_name == marker_2:
                        last_index = j
                        break

                if (first_index == -1) or (last_index == -1):
                    raise DatabaseRequestException("Didn't find two different markers. Shouldn't be possible. " +
                                                   "Please excuse this error and forward it to me.")

                # wrong order -> nothing to do...
                if last_index < first_index:
                    continue

                try:
                    sorted_dict[well_id][first_index].session = session
                    point = Requests.create_thickness_point(sorted_dict, well_id, first_index, last_index, session,
                                                            use_faulted, fault_name, (('summarised', 'bool', 1),))
                    geopoints.append(point)
                # FaultException -> do nothing except catching the exception
                except FaultException:
                    continue

                # finished summarise section -> continue to avoid second round without summarise
                continue
            # don't summarise
            first_index = -1
            for index in range(len(sorted_dict[well_id])):
                if sorted_dict[well_id][index].horizon.unit_name == marker_1:
                    first_index = index
                elif (first_index != -1) and (sorted_dict[well_id][index].horizon.unit_name == marker_2):
                    try:
                        sorted_dict[well_id][first_index].session = session
                        point = Requests.create_thickness_point(sorted_dict, well_id, first_index, index, session,
                                                                use_faulted, fault_name,
                                                                (('multiple marker', 'bool', 1),))
                        geopoints.append(point)
                    # FaultException -> do nothing except catching the exception
                    except FaultException:
                        continue
                    finally:
                        first_index = -1

        return geopoints

    @staticmethod
    def interpolate_geopoints(points, property_name, method):
        # type: (List[GeoPoint], str, str) -> None
        """
        Interpolate the property values of the given GeoPoints using the interpolation method 'method'

        .. todo:: - Integrate functionality
                  - define interpolation methods
                  - define return statement

        :param points: List of GeoPoints as interpolation base
        :type points: List[GeoPoint]

        :param property_name: Name of the property to interpolate
        :type property_name: str

        :param method: Interpolation method
        :type method: str

        :return: Currently Nothing
        :raises TypeError: if on of the points is not of type GeoPoint

        possible values for interpolation **method** are:

        - **nearest** (Nearest Neighbour interpolation)
        - **ide** (Inverse Distance interpolation)
        - **spline** (Thin-Plate-Spline interpolation)
        """
